/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>

#include "SortWdgt.h"

#include "Sort.h"
#include "BubbleSort.h"
#include "InsertionSort.h"
#include "MergeSort.h"
#include "QuickSort.h"
#include "SelectionSort.h"
#include "ShellSort.h"

#include "sort.xbm"

#include <ordcltn.h>

XtAppContext app;
#define NAM "SortingInAction"

#define NUM_DATA 50
#define DATA_MAXVAL 40

int sortcnt;
int busy;
int delay;
int delaypid;
int deadlockflag;
id cls[NUM_SORT_TYPES];
id sorttab[NUM_SORT_TYPES];
int pidtab[NUM_SORT_TYPES];
int fintab[NUM_SORT_TYPES];
int ticktab[NUM_SORT_TYPES];
Widget sortwindow;
int neednewdata;
int needsortwindow;
Widget sortwidgets[NUM_SORT_TYPES];
Widget toggle[NUM_SORT_TYPES];
Widget ticklabel;
XtIntervalId ticktimer;

void setlabel(Widget w,const char *msg)
{
  int n = 0;
  Arg args[2];
  XtSetArg(args[n],XtNlabel,msg);	n++;
  XtSetValues(w,args,n);
}

void setenabled(Widget w,BOOL flag)
{
  int n = 0;
  Arg args[2];
  XtSetArg(args[n],XtNstate,&flag);	n++;
  XtSetValues(w,args,n);
}

BOOL isenabled(Widget w)
{
  int n = 0;
  BOOL flag;
  Arg args[2];
  XtSetArg(args[n],XtNstate,&flag);	n++;
  XtGetValues(w,args,n);
  return flag;
}

void genrandata(int size,int max,int percent)
{
  int c;
  static int randata[NUM_DATA +1];

  if (neednewdata) {
    srandom(time(0));

    for (c = 0; c < size ;c++) {
      if (((random() % 100) + 1) > percent)
	 randata[c] = (random() % (int)(max-1))+ 1;
      else
	 randata[c] = (int)(c*(max/(float)size)) +1;
    }
  }

  for(c=0;c<sortcnt;c++) {
    setbarlabel(sortwidgets[c],[sorttab[c] getName]);
    setbardata(sortwidgets[c],randata,size,DATA_MAXVAL);
    assert(sorttab[c]);
    [sorttab[c] setsize:size address:randata];
  }
}

void setanimate(BOOL flag)
{
  int i;
  for(i=0;i<sortcnt;i++) {
    [sorttab[i] setanimate:flag];
  }
}

int numticks(void)
{
  int i;
  int max = 0;
  for(i=0;i<sortcnt;i++) {
    if (ticktab[i] > max) max = ticktab[i];
  }
  return max;
}

void markasdone(int pid)
{
  int i;
  for(i=0;i<sortcnt;i++) {
    if (pidtab[i] == pid) {
      assert(fintab[i] == 0);
      fintab[i] = 1;
      return;
    }
  }
  fprintf(stderr,"Unknown child %d\n",pid);
}

void delayedwakeup(XtPointer clientdata,XtIntervalId *timer)
{
  int c = (int)clientdata;
  if (fintab[c] == 0) [sorttab[c] wakeup];
}

void schedchild(void)
{
  if (busy) {
    int c = 0;
    int i,min; 

    for(i=0;i<sortcnt;i++) {
      if (sorttab[i] != nil && fintab[i] == 0) {
	c = i;
	min = ticktab[i];
	break;
      }
    }
    for(;i<sortcnt;i++) {
      if (sorttab[i] != nil && fintab[i] == 0 && ticktab[i] < min) { 
	c = i;
	min = ticktab[i];
      }
    }

    if (delay) {
      XtAppAddTimeOut(app,delay,delayedwakeup,(XtPointer)c);
    } else {
      [sorttab[c] wakeup];
    }
  }
}

void avoiddeadlock(void)
{
  int i,c = 0;

  deadlockflag = 1;

  for(i=0;i<sortcnt;i++) {
    if (fintab[i] == 0) c++;
  }

  if (c == 0) busy = 0;
}

void childdied(int signo,int code,struct sigcontext *ctxt) 
{
  int pid;
  int status;
  while ((pid = waitpid(0,&status,WNOHANG)) > 0) {
    markasdone(pid);
  }
  avoiddeadlock();
}

void startsort(void)
{
  int i;

  deadlockflag = 0;
  busy = 1;

  for(i=0;i<sortcnt;i++) {
    if (sorttab[i]) {
      fintab[i]=0;
      ticktab[i]=0;
      pidtab[i]=[sorttab[i] fork];
    } else {
      pidtab[i]=0;
      ticktab[i]=0;
      fintab[i]=1;
    }
  }

  /* avoid scheduling deadlock in case only one sort selected */
  /* (case sortcnt == 1) */
  avoiddeadlock();
}

void readfrom(XtPointer clientdata,int *source,XtInputId *input)
{
  int i;
  id sort;
  sortmsg m;

  i = (int)clientdata;
  sort = sorttab[i];

  if ([sort receive:&m]) {
    Widget w = sortwidgets[i];
    int p = m.position[0];
    int q = m.position[1];
    int a = m.value[0];
    int b = m.value[1];
    switch(m.what) {
      case SORTING_MSG: {
	break;
      }
      case COMPARE_MSG: {
	comparebars(w,p,a,q,b);
	break;
      }
      case SWAP_MSG: {
	swapbars(w,p,a,q,b);
	break;
      }
      case MOVE_MSG: {
	movebar(w,a,p,b);
	break;
      }
      case TICK_MSG: {
        ticktab[i] = a;
	schedchild();
	break;
      }
      case FINISHED_MSG: {
	char buf[100];
	sprintf(buf,"%s (%i Ticks)",[m.who getName],numticks());
	setbarlabel(w,buf);
	break;
      }
    }
  } else {
    XtRemoveInput(*input);
    sorttab[i] = [sort free]; /* close parent end */
  }
}

void monitorpipes(void)
{
  int i;
  for(i=0;i<sortcnt;i++) {
    id sort = sorttab[i];
    if (sort) {
      XtPointer cond = (XtPointer)XtInputReadMask;
      XtAppAddInput(app,[sort drawfd],cond,readfrom,(XtPointer)i);
    }
  }
}

void setupcls(void)
{
  cls[0] = [BubbleSort class];
  cls[1] = [InsertionSort class];
  cls[2] = [MergeSort class];
  cls[3] = [QuickSort class];
  cls[4] = [SelectionSort class];
  cls[5] = [ShellSort class];
  assert(5 < NUM_SORT_TYPES);
}

void fillsorttab(void)
{
  int i;
  setupcls();
  for(sortcnt = 0,i = 0;i<NUM_SORT_TYPES;i++) {
    if (isenabled(toggle[i])) sorttab[sortcnt++] = [cls[i] new];
  }
}

void setticklabel(XtPointer clientdata,XtIntervalId *timer)
{
  char buf[100];
#define TICK_INTERVAL 100

  sprintf(buf,"%d T",numticks());
  setlabel(ticklabel,buf);
  
  if (busy) {
    /* reinstall as long as busy */
    ticktimer = XtAppAddTimeOut(app,TICK_INTERVAL,setticklabel,NULL);
  } else {
    ticktimer = (XtIntervalId)NULL;
  }

  if (deadlockflag) {
    deadlockflag = 0;
    schedchild();
  }
}

void starttickcounter(void)
{
  ticktimer = XtAppAddTimeOut(app,TICK_INTERVAL,setticklabel,NULL);
}

void stoptickcounter(void)
{
  XtRemoveTimeOut(ticktimer);
  ticktimer = (XtIntervalId)NULL;
}

void tracktoggle(Widget w,XtPointer clientdata,XtPointer calldata)
{
  needsortwindow = 1;
}

void sortbuttons(Widget toplevel)
{
  int i;
  Widget w;

  toplevel = XtCreateManagedWidget("box2",boxWidgetClass,toplevel,NULL,0);
  
  setupcls();
  for(i=0;i<NUM_SORT_TYPES;i++) {
    w=XtCreateManagedWidget([cls[i] name],toggleWidgetClass,toplevel,NULL,0);
    XtAddCallback(w,XtNcallback,tracktoggle,NULL);
    setenabled(w,YES);
    toggle[i]=w;
  }
}

void seticonifyicon(Widget toplevel)
{
  int n = 0;
  Pixmap icon;
  Arg args[1];
  
  icon = XCreateBitmapFromData(XtDisplay(toplevel),XtWindow(toplevel), sort_bits, sort_width, sort_height);

  XtSetArg(args[n], XtNiconPixmap, icon);	n++;
  XtSetValues(toplevel, args, n);
} 

void animationwidgets(Widget toplevel)
{
  int i;
  toplevel = XtCreateManagedWidget("frame2",boxWidgetClass,toplevel,NULL,0);
  for (i=0;i<sortcnt;i++) {
    Widget w;
    char buf[10];
    sprintf(buf,"s%i",i);
    w = XtCreateManagedWidget(buf,sortWidgetClass,toplevel,NULL,0);
    sortwidgets[i] = w;
  }
}  

void newsortwindow(Display *dp)
{
  if (sortwindow) {
    XtUnrealizeWidget(sortwindow); 
    XtDestroyWidget(sortwindow);
  }
  sortwindow = XtAppCreateShell(NULL,NULL,topLevelShellWidgetClass,dp,NULL,0);
  fillsorttab();
  animationwidgets(sortwindow);
  neednewdata = 1;
  genrandata(NUM_DATA,DATA_MAXVAL,0);
  XtRealizeWidget(sortwindow);
  seticonifyicon(sortwindow);
  needsortwindow = 0;
  neednewdata = 0;
}

void gosort(Widget w,XtPointer clientdata,XtPointer calldata)
{
  if (busy) return;

  if (needsortwindow) {
    newsortwindow(XtDisplay(w));
  } else {
    fillsorttab();
    setanimate(YES);
    genrandata(NUM_DATA,DATA_MAXVAL,0);
    starttickcounter();
    startsort();
    monitorpipes();
    neednewdata = 1;
  }
}

void quit(Widget w,XtPointer clientdata,XtPointer calldata)
{
  kill(0,SIGTERM);
}

void abortsorts(void)
{
  int i;
  for(i=0;i<sortcnt;i++) {
    if (fintab[i] == 0) [sorttab[i] abort];
  }
}

void stopsort(Widget w,XtPointer clientdata,XtPointer calldata)
{
  if (busy) {
    abortsorts();
  }
}

void faster(Widget w,XtPointer clientdata,XtPointer calldata)
{
  delay = 0;
}

void normal(Widget w,XtPointer clientdata,XtPointer calldata)
{
  delay = 10;
}

void slower(Widget w,XtPointer clientdata,XtPointer calldata)
{
  delay = 500;
}

void iconbuttons(Widget toplevel)
{
  Widget w;
  toplevel = XtCreateManagedWidget("bx1",boxWidgetClass,toplevel,NULL,0);
  ticklabel = XtCreateManagedWidget("Ticks",labelWidgetClass,toplevel,NULL,0);
  w=XtCreateManagedWidget("Go !",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,gosort,NULL);
  w=XtCreateManagedWidget("Stop !",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,stopsort,NULL);
  w=XtCreateManagedWidget("Slow",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,slower,NULL);
  w=XtCreateManagedWidget("Normal",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,normal,NULL);
  w=XtCreateManagedWidget("Fast",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,faster,NULL);
  w=XtCreateManagedWidget("Quit",commandWidgetClass,toplevel,NULL,0);
  XtAddCallback(w,XtNcallback,quit,NULL);
}

void sortparameters(Widget toplevel)
{
  toplevel = XtCreateManagedWidget("frame1",boxWidgetClass,toplevel,NULL,0);
  sortbuttons(toplevel);
}  

void main(int argc,char **argv)
{
  Display *dp;
  Widget w1,w2;
  struct sigaction na,oa;

  na.sa_handler = childdied;
  sigemptyset(&na.sa_mask);
  na.sa_flags = SA_RESTART;
  if (sigaction(SIGCHLD,&na,&oa) < 0) {
    perror("sigaction");exit(1);
  }

  delay = 50;

  w1 = XtAppInitialize(&app,NAM,NULL,0,&argc,argv,NULL,NULL,0);
  dp = XtDisplay(w1);
  w2 = XtAppCreateShell(NULL,NULL,topLevelShellWidgetClass,dp,NULL,0);

  sortparameters(w1);
  XtRealizeWidget(w1);
  seticonifyicon(w1);

  iconbuttons(w2);
  XtRealizeWidget(w2);
  seticonifyicon(w2);

  newsortwindow(dp);

  XtAppMainLoop(app);
}

